package com.sample;

import bitronix.tm.BitronixTransactionManager;
import bitronix.tm.TransactionManagerServices;
import bitronix.tm.resource.jdbc.PoolingDataSource;
import java.io.File;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletionService;
import java.util.concurrent.Executor;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;
import javax.transaction.SystemException;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import org.drools.ClockType;
import org.drools.KnowledgeBase;
import org.drools.KnowledgeBaseConfiguration;
import org.drools.KnowledgeBaseFactory;
import org.drools.builder.KnowledgeBuilder;
import org.drools.builder.KnowledgeBuilderFactory;
import org.drools.builder.ResourceType;
import org.drools.conf.EventProcessingOption;
import org.drools.event.rule.ObjectInsertedEvent;
import org.drools.event.rule.ObjectUpdatedEvent;
import org.drools.event.rule.ObjectRetractedEvent;
import org.drools.event.rule.WorkingMemoryEventListener;
import org.drools.impl.EnvironmentFactory;
import org.drools.io.ResourceFactory;
import org.drools.persistence.jpa.JPAKnowledgeService;
import org.drools.runtime.Environment;
import org.drools.runtime.EnvironmentName;
import org.drools.runtime.KnowledgeSessionConfiguration;
import org.drools.runtime.StatefulKnowledgeSession;
import org.drools.runtime.conf.ClockTypeOption;
import org.drools.runtime.rule.WorkingMemoryEntryPoint;
import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

/**
 * Tests intermittent errors during serialization under high load.
 */
public class CEPPersistenceTest {

    private static EntityManagerFactory emf;

    private static PoolingDataSource ds;

    private static BitronixTransactionManager tm;

    @BeforeClass
    public static void setup() throws SystemException {
        ds = setupDataSource();

        Map<String, String> map = new HashMap<String, String>();
        emf = Persistence.createEntityManagerFactory("org.drools.persistence.jpa", map);

        tm = TransactionManagerServices.getTransactionManager();
        tm.setTransactionTimeout(3600); // longer timeout for a debugger
    }

    @AfterClass
    public static void cleanup() {
        emf.close();
        TransactionManagerServices.getTransactionManager().shutdown();
        ds.close();
    }

    public static PoolingDataSource setupDataSource() {
        // Please edit here when you want to use your database
        PoolingDataSource pds = new PoolingDataSource();
        pds.setUniqueName("jdbc/jbpm-ds");
        pds.setClassName("bitronix.tm.resource.jdbc.lrc.LrcXADataSource");
        pds.setMaxPoolSize(5);
        pds.setAllowLocalTransactions(true);
        pds.getDriverProperties().put("driverClassName", "org.hsqldb.jdbcDriver");
        pds.getDriverProperties().put("url", "jdbc:hsqldb:file:" + new File("target/bitronixDB").getAbsolutePath());
        pds.init();
        return pds;
    }

    public static StatefulKnowledgeSession newStatefulKnowledgeSession(KnowledgeBase kbase) {
        return loadStatefulKnowledgeSession(kbase, -1);
    }

    public static StatefulKnowledgeSession loadStatefulKnowledgeSession(KnowledgeBase kbase, int sessionId) {
        StatefulKnowledgeSession ksession;
        Environment env = getEnv();

        KnowledgeSessionConfiguration ksconf = KnowledgeBaseFactory.newKnowledgeSessionConfiguration();
        ksconf.setOption(ClockTypeOption.get(ClockType.REALTIME_CLOCK.getId()));

        if (sessionId == -1) {
            ksession = JPAKnowledgeService.newStatefulKnowledgeSession(kbase, ksconf, env);
        } else {
            ksession = JPAKnowledgeService.loadStatefulKnowledgeSession(sessionId, kbase, ksconf, env);
        }

        return ksession;
    }

    private static Environment getEnv() {
        Environment env = EnvironmentFactory.newEnvironment();
        env.set(EnvironmentName.ENTITY_MANAGER_FACTORY, emf);
        env.set(EnvironmentName.TRANSACTION_MANAGER, TransactionManagerServices.getTransactionManager());
        return env;
    }

    @Test
    public void testParallelInsertsOnPersistedSession() throws Exception {
        String str
                = "import com.sample.CEPPersistenceTest.ParkingEvent;\n"
                + "\n"
                + "declare ParkingEvent\n"
                + " @role( event )\n"
                + " @expires(1s)\n"
                + "end\n"
                + "\n"
                + "rule \"Hello Event\"\n"
                + " when\n"
                + " $p : ParkingEvent( ) from entry-point \"parking\"\n"
                + " then\n"
                + //" System.out.println( $p.toString() );\n" +
                "end\n";

        KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder();
        kbuilder.add(ResourceFactory.newByteArrayResource(str.getBytes()),
                ResourceType.DRL);
        KnowledgeBaseConfiguration conf = KnowledgeBaseFactory.newKnowledgeBaseConfiguration();
        conf.setOption(EventProcessingOption.STREAM);
        KnowledgeBase kbase = KnowledgeBaseFactory.newKnowledgeBase(conf);

        if (kbuilder.hasErrors()) {
            fail(kbuilder.getErrors().toString());
        }

        kbase.addKnowledgePackages(kbuilder.getKnowledgePackages());

        KnowledgeSessionConfiguration ksconf = KnowledgeBaseFactory.newKnowledgeSessionConfiguration();
        ksconf.setOption(ClockTypeOption.get(ClockType.REALTIME_CLOCK.getId()));
        final StatefulKnowledgeSession ksession = JPAKnowledgeService.newStatefulKnowledgeSession(kbase, ksconf, getEnv());

        ksession.addEventListener(new WorkingMemoryEventListener() {
            @Override
            public void objectUpdated(ObjectUpdatedEvent arg0) {
                ksession.fireAllRules();
            }

            @Override
            public void objectRetracted(ObjectRetractedEvent arg0) {
                ksession.fireAllRules();
            }

            @Override
            public void objectInserted(ObjectInsertedEvent arg0) {
                ksession.fireAllRules();
            }
        });

        assertNull(ksession.getWorkingMemoryEntryPoint("notExisting"));

        WorkingMemoryEntryPoint ep = ksession.getWorkingMemoryEntryPoint("parking");

        Executor executor = Executors.newCachedThreadPool(new ThreadFactory() {
            public Thread newThread(Runnable r) {
                Thread t = new Thread(r);
                t.setDaemon(true);
                return t;
            }
        });

        final int THREAD_NR = 30;
        CompletionService<Boolean> ecs = new ExecutorCompletionService<Boolean>(executor);
        for (int i = 0; i < THREAD_NR; i++) {
            TestEventRunner runner = new TestEventRunner();
            runner.setEp(ep);
            runner.setKsession(ksession);
            ecs.submit(runner);
        }

        // fireUntilHalt() not supported with persisted session anymore
        // ksession.fireUntilHalt();
        boolean success = true;
        for (int i = 0; i < THREAD_NR; i++) {
            try {
                success = ecs.take().get() && success;
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }

        assertTrue(success);

        // let events expire
        try {
            Thread.sleep(10000L);
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }

        tm.begin();
        try {
            ksession.fireAllRules();
            tm.commit();
        } catch (Exception e) {
            try {
                tm.rollback();
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }
        assertEquals(0, ep.getFactCount());

        // ksession.halt();
        ksession.dispose();
    }

    public static class TestEventRunner implements Callable<Boolean> {

        private WorkingMemoryEntryPoint ep;

        private StatefulKnowledgeSession ksession;

        public StatefulKnowledgeSession getKsession() {
            return ksession;
        }

        public void setKsession(StatefulKnowledgeSession ksession) {
            this.ksession = ksession;
        }

        public WorkingMemoryEntryPoint getEp() {
            return ep;
        }

        public void setEp(WorkingMemoryEntryPoint ep) {
            this.ep = ep;
        }

        public Boolean call() {
            try {
                Thread.sleep(500); // make sure ksession started

                tm.begin();        // with P05, a transaction is needed here!
                for (int i = 0; i < 5; i++) {
                    ep.insert(new ParkingEvent("CarA", "ParkingA", ParkingEvent.ENTER));
                    Thread.sleep(200);
                    ep.insert(new ParkingEvent("CarB", "ParkingA", ParkingEvent.ENTER));
                    Thread.sleep(200);
                    ep.insert(new ParkingEvent("CarA", "ParkingA", ParkingEvent.EXIT));
                    Thread.sleep(200);
                    ep.insert(new ParkingEvent("CarA", "ParkingB", ParkingEvent.ENTER));
                    Thread.sleep(200);
                }

                tm.commit();
            } catch (Exception e) {
                try {
                    tm.rollback();
                } catch (Exception ex) {
                    e.printStackTrace();
                }
                e.printStackTrace();
                return false;
            }
            return true;
        }
    }

    public static class ParkingEvent implements Serializable {

        public static final int ENTER = 0;
        public static final int EXIT = 1;

        private String car;

        private String zone;

        private int type;

        public ParkingEvent() {

        }

        public ParkingEvent(String car, String zone, int type) {
            this.car = car;
            this.zone = zone;
            this.type = type;
        }

        public String getCar() {
            return car;
        }

        public void setCar(String car) {
            this.car = car;
        }

        public String getZone() {
            return zone;
        }

        public void setZone(String zone) {
            this.zone = zone;
        }

        public int getType() {
            return type;
        }

        public void setType(int type) {
            this.type = type;
        }

        @Override
        public String toString() {
            return "ParkingEvent : car = " + car + ", zone = " + zone + ", type = " + type;
        }
    }

}
